;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "citre" "20260530.602"
  "Superior code reading & auto-completion tool with pluggable backends."
  '((emacs "26.1"))
  :url "https://github.com/universal-ctags/citre"
  :commit "be0d9c6dc9b1ac67d76fc7ed315f2369d5c3bde8"
  :revdesc "be0d9c6dc9b1"
  :keywords '("convenience" "tools")
  :authors '(("Hao Wang" . "amaikinono@gmail.com"))
  :maintainers '(("Hao Wang" . "amaikinono@gmail.com")))
